Program RAYPAS;

{
  Sample Raytracing program
  Copyright (C) 1998  Paul Toth

  TothPaul@multimania.com
  http://www.multimania.com/tothpaul

  Thanks to Cyrille FAVREAU for is Delphi RayTracer !
  http://www.multimania.com/cfavreau

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

{$N+,E-}

Const
 BigZ=10000000;

Type
 PPrimitive=^TPrimitive;

 TPoint=record x,y,z:single end;

 TRGBColor=record R,G,B:word end;

 TPixel=record
  Owner :PPrimitive;
  Dist  :single;
  Color :TRGBColor;
  Point :TPoint;
  Inter :TPoint;
  Normal:TPoint;
  Reflection:TPoint;
 end;

 TVesaInfo=record
  VESA         : array[1..4] of char;
  Version      : word;
  OEMName      : PChar;
  Capabilities : longint;
  VESAModes    : pointer;
  Memory       : word;
  Reserved     : array[20..255] of byte;
 end;

 TVesaMode=record
  Flags         :set of (fSupported,fMore,fBIOS,fColor,fGraphic);
  HiFlags       :byte;
  WindowAFlags  ,
  WindowBFlags  :set of (fAvail,fRead,fWrite);
  Granularity   :word;  { Bank*Granularity = Video Offset }
  WindowSize    :word;  { Allways 64K ? }
  WindowASegment:word;  { Allways $A000 ? }
  WindowBSegment:word;
  SetBank       :pointer;
  BytesPerline  :word;
 {-More-}
  Width         :word;
  Height        :word;
  CharWidth     :byte;
  CharHeight    :byte;
  PlaneCount    :byte;
  BitsPerPixel  :byte;
  BankCount     :byte;
  MemoryModel   :byte;
  BankSize      :byte;
  Reserved      :array[29..256] of byte;
 end;

 TRender=object
  Width,Height:integer;
  Constructor Create;
  Destructor Destroy;
  procedure PutPixel(x,y,r,g,b:word);
  procedure Clear;
  procedure Line(x1,y1,x2,y2,r,g,b:word);
  procedure Circle(x,y,rd,r,g,b:word);
 end;

 TPrimitive=object
  Next:PPrimitive;
  Center:TPoint;
  Radius:single;
  Color:TRGBColor;
  Brightness:single;
  Transparency:single;
  function Intersection(var Src,Dst:TPoint;Var Pixel:TPixel):boolean; virtual;
  procedure DrawXZ(Var ARender:TRender); virtual;
  procedure Reflect(Var Pixel:TPixel);
 end;

 TScene=object
  Eye:TPoint;
  Ray:TPoint;
  Back:TRGBColor;
  Lamps:PPrimitive;
  Objects:PPrimitive;
  Pixel:TPixel;
  constructor Create;
  procedure DrawXZ(Var ARender:TRender);
  procedure Draw3D(Var ARender:TRender);
  Function FindIntersection(var Src,Dst:TPoint; var APixel:TPixel):boolean;
  procedure PixelColor(Var Src:TPoint;Var APixel:TPixel;Generation:word);
 end;

 TLamp=object(TPrimitive)
  constructor Create(Var AScene:TScene);
  procedure DrawXZ(Var ARender:TRender); virtual;
 end;

 TSphere=object(TPrimitive)
  constructor Create(Var AScene:TScene);
  function Intersection(var Src,Dst:TPoint;Var Pixel:TPixel):boolean; virtual;
  procedure DrawXZ(Var ARender:TRender); virtual;
 end;

Procedure Error(Msg:string);
 begin
  writeln(msg);
  halt;
 end;

Procedure Normalize(Var N:TPoint);
 var
  l:single;
 begin
  with N do begin
   l:=sqrt(sqr(x)+sqr(y)+sqr(z));
   if l<>0 then begin
    x:=x/l;
    y:=y/l;
    z:=z/l;
   end;
  end;
 end;

Function NormalDot(Var A,B:TPoint):single;
 begin
  NormalDot:=(A.X*B.X)+(A.Y*B.Y)+(A.Z*B.Z);
 end;

Procedure Reflect(Var Incident,Normal,Reflected:TPoint);
 var
  n:TPoint;
  d:single;
 begin
  n:=Incident;
  Normalize(n);
  d:=NormalDot(N,Normal);
  Reflected.x:=Incident.x-d*Normal.x;
  Reflected.y:=Incident.y-d*Normal.y;
  Reflected.z:=Incident.z-d*Normal.z;
 end;

 (*
 Procedure Reflect(aIncident,aNormal: TCoord; Var aReflected: TCoord);
 Var
   d: Extended;
 Begin
  d := 2*StdNormalDot(aIncident,aNormal);
  aReflected.x := aIncident.x - d*aNormal.x;
  aReflected.y := aIncident.y - d*aNormal.y;
  aReflected.z := aIncident.z - d*aNormal.z;
 End; // Reflect
 *)

Function GetVesaInfo(Var VesaInfo:TVesaInfo):boolean; assembler;
 asm
  mov ax,4F00h
  les di,VesaInfo
  int 10h
  cmp al,004Fh
  je  @ok
   mov al,False
  @ok:
 end;

Function GetModeInfo(Mode:Word; Var VesaMode:TVesaMode):boolean; assembler;
  asm
   mov ax,$4F01
   mov cx,Mode
   les di,VesaMode
   int 10h
   cmp al,004Fh
   je  @ok
    mov al,False
   @Ok:
  end;

Function Hexa(w,size:word):string;
 const
  hex:array[0..$F] of char='0123456789ABCDEF';
 var
  result:string;
 begin
  result:='';
  repeat
   result:=hex[w and $F]+Result;
   w:=w shr 4;
  until w=0;
  while length(result)<size do result:='0'+result;
  Hexa:=Result;
 end;

Procedure ShowVesaModes;
 var
  VesaInfo:TVesaInfo;
  Mode:^Word;
  ModeInfo:TVesaMode;
 begin
  if not GetVesaInfo(VesaInfo) then Error('Not a VESA card');
  WriteLn(#16' ',VesaInfo.OEMName,' (Ver ',hi(VesaInfo.Version),'.',lo(VesaInfo.Version),') ',VesaInfo.Memory*64,'Ko');
  Mode:=VesaInfo.VesaModes;
  while mode^<>$FFFF do begin

   write(' mode ',hexa(mode^,3),'h at');

   If GetModeInfo(mode^,ModeInfo) then begin

    if fSupported in ModeInfo.Flags then begin

     if fAvail in ModeInfo.WindowAFlags then begin
      Write(' $',Hexa(ModeInfo.WindowASegment,4));
     end;

     if fAvail in ModeInfo.WindowBFlags then begin
      Write(' $',Hexa(ModeInfo.WindowASegment,4));
     end;

     Write(' Window=',ModeInfo.WindowSize,'K/',ModeInfo.Granularity:2,'K banks - ');

     if fGraphic in ModeInfo.Flags then write('Graphic ') else  write('Text    ');

     if fMore in ModeInfo.Flags then
      Write(ModeInfo.Width,'x',ModeInfo.Height,'x',longint(1) shl ModeInfo.BitsPerPixel)
     else
      if fColor in ModeInfo.Flags then write('color') else write('b&w');

     writeln;

    end else
     writeln('unsupported ???');

   end else
    WriteLn('unavailable ???');

   inc(mode);
  end;
 end;

Function SetMode(Mode:word):boolean; assembler;
 asm
  mov ax,$4F02
  mov bx,Mode
  int 10h
  cmp al,004Fh
  je @ok
   mov al,False
  @ok:
 end;

var
 VesaBank:pointer;
 VesaWin :word;
 VesaSeg :word;
 BankSize:word;
 LineSize:word;

Function SetBank(ABank:word):boolean; assembler;
 asm
  mov  bx,VesaWin
  mov  ax,ABank
  mul  BankSize
  mov  dx,ax
  call VesaBank
  cmp  ax,004Fh
  je   @ok
   mov al,False
  @ok:
 end;

Constructor TRender.Create;
 Const
  Mode=$112;
 var
  VesaInfo:TVesaInfo;
  VesaMode:TVesaMode;
 begin
  if not GetVesaInfo(VesaInfo) then Error('Not a VESA card');
  if not GetModeInfo(Mode,VesaMode) then Error('Unsupported mode 112h');
  if VesaMode.BitsPerPixel<>24 then Error('Mode 112h isn''t a 16M color mode');

  Width :=VesaMode.Width;
  Height:=VesaMode.Height;

  VesaBank:=VesaMode.SetBank;
  BankSize:=64 div VesaMode.Granularity;
  LineSize:=VesaMode.BytesPerLine;

  if fAvail in VesaMode.WindowAFlags then begin
   VesaWin :=0;
   VesaSeg :=VesaMode.WindowASegment;
  end else begin
   if not (fAvail in VesaMode.WindowBFlags) then Error('Not VesaWin found ???');
   VesaWin :=1;
   VesaSeg :=VesaMode.WindowBSegment;
  end;

  if not SetMode(Mode) then Error('Unable to set VESA Mode 112h');
 end;

Destructor TRender.Destroy;
 begin
  asm mov ax,03h; int 10h end;
 end;

Procedure TRender.Putpixel(x,y,r,g,b:word);
 var
  bank  :word;
  pixel :word;
 begin
  asm
   xor dx,dx
   mov ax,y
   mul LineSize
   mov bx,x
   mov cx,bx
   add bx,bx
   add bx,cx
   add ax,bx
   adc dx,0
   mov pixel,ax
   mov bank,dx
  end;
  SetBank(Bank);
  Mem[VesaSeg:Pixel]:=B; inc(Pixel);
  Mem[VesaSeg:Pixel]:=G; inc(Pixel);
  Mem[VesaSeg:Pixel]:=R;
 end;

Procedure TRender.Clear;
 var
  x,y:word;
 begin
  for x:=0 to 14 { 3*640*480/64K } do begin
   SetBank(x);
  { fillchar(Mem[VesaSeg:0],$FFFF,255); }
   asm
    mov es,VesaSeg
    xor di,di
   { !!!!!!!! LineSize<>3*Width :(( !!!!!!!!!!
    mov cx,$FFFF/3
   @nextpixel:
    mov ax,B; stosb
    mov ax,G; stosb
    mov ax,R; stosb
   loop @nextpixel
   end;
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!}
    mov cx,$FFFF
    mov al,0
    rep stosb
   end;
  end;

{  for y:=0 to Height-1 do
   for x:=0 to Width-1 do PutPixel(x,y,R,G,B); }
 end;

procedure TRender.Line(x1,y1,x2,y2,r,g,b:word);
 var
  dx,dy:integer;
  ix,iy:integer;
  ex,ey:integer;
  ee,ii:integer;
 begin
  dx:=x2-x1; if dx<0 then begin dx:=-dx; ix:=-1 end else ix:=+1;
  dy:=y2-y1; if dy<0 then begin dy:=-dy; iy:=-1 end else iy:=+1;
  ex:=dx;
  ey:=dy;
  if ex<ey then ee:=ey else ee:=ex;
  for ii:=0 to ee do begin
   if (x1>=0)and(x1<Width)
   and(y1>=0)and(y1<Height) then PutPixel(x1,y1,r,g,b);
   inc(ex,dx);
   if ex>ee then begin
    dec(ex,ee);
    inc(x1,ix);
   end;
   inc(ey,dy);
   if ey>ee then begin
    dec(ey,ee);
    inc(y1,iy);
   end;
  end;
 end;

procedure TRender.Circle(x,y,rd,r,g,b:word);
 var
  a:byte;
  ax,ay:integer;
  lx,ly:integer;
 begin
  lx:=rd;
  ly:=0;
  for a:=1 to 16 do begin
   ax:=round(rd*cos(a*PI/32));
   ay:=round(rd*sin(a*PI/32));
   line(x+lx,y+ly,x+ax,y+ay,r,g,b);
   line(x-lx,y+ly,x-ax,y+ay,r,g,b);
   line(x+lx,y-ly,x+ax,y-ay,r,g,b);
   line(x-lx,y-ly,x-ax,y-ay,r,g,b);
   lx:=ax;
   ly:=ay;
  end;
 end;

Function TPrimitive.Intersection(var Src,Dst:TPoint;Var Pixel:TPixel):boolean;
 begin
  Intersection:=false;
 end;

Procedure TPrimitive.DrawXZ(Var ARender:TRender);
 var
  cx,cy:integer;
 begin
  cx:=ARender.Width div 2+Round(Center.X);
  cy:=ARender.Height div 2-Round(Center.Z);
  ARender.Line(cx-5,cy,cx+5,cy,Color.R,Color.G,Color.B);
  ARender.Line(cx,cy-5,cx,cy+5,Color.R,Color.G,Color.B);
 end;

Procedure TPrimitive.Reflect(Var Pixel:TPixel);
 begin
 (*
      // ================================================ //
      // Calculate the reflected vector                   //
      //                                                  //
      //                  ^ Normal to object surface (N)  //
      //     Reflection1  |                               //
      //                \ |  Eye (O_I)                    //
      //                 \| /                             //
      //  ----------------O--------------- Object surface //
      //             Intersection                         //
      //                                                  //
      // ================================================ //
      O_I := TCoord.CreateCoord(Origin.x-ClosestIntersection.x,Origin.y-ClosestIntersection.y,Origin.z-ClosestIntersection.z);
      N   := TCoord.CreateCopyCoord(TPrimitive(ObjectList.Items[ClosestObject]).GetNormalVector(ClosestIntersection));
      Reflect(O_I,N,Reflection1);
      N.Free;
 *)
 { Normalize Pixel }
 With Pixel do begin
  with Point do begin
   Dist:=sqrt(sqr(x)+sqr(y)+sqr(z));
   if Dist<>0 then begin
    x:=x/Dist;
    y:=y/Dist;
    z:=z/Dist;
   end;
  end;
  With Normal do begin
   Dist:=sqrt(sqr(x)+sqr(y)+sqr(z));
   if Dist<>0 then begin
    x:=x/Dist;
    y:=y/Dist;
    z:=z/Dist;
   end;
   Dist:=(x*Point.x)+(y*Point.y)+(z*Point.z);
  end;
 end;
 end;

Constructor TScene.Create;
 begin
  Objects:=nil;
  Lamps:=nil;

  Eye.x:=0;
  Eye.y:=0;
  Eye.z:=-480;

  Back.R:=64;
  Back.G:=64;
  Back.B:=64;
 end;

Procedure TScene.DrawXZ(Var ARender:TRender);
 var
  cx,cy:integer;
  obj:PPrimitive;
 begin
  cx:=ARender.Width div 2;
  cy:=ARender.Height div 2;
 { clear }
  ARender.Clear;
 { axes }
  ARender.Line(0,cy,ARender.Width-1,cy,0,0,255);
  ARender.Line(cx,0,cx,ARender.Height-1,0,0,255);
 { Camera }
  ARender.Line(0,cy,cx,cy-round(Eye.z),0,255,0);
  ARender.Line(ARender.Width-1,cy,cx,cy-Round(Eye.z),0,255,0);
 { Lamps }
  obj:=Lamps;
  while obj<>nil do begin
   obj^.DrawXZ(ARender);
   obj:=obj^.Next;
  end;
 { Objects }
  obj:=Objects;
  while obj<>nil do begin
   obj^.DrawXZ(ARender);
   obj:=obj^.Next;
  end;
 end;

Procedure TScene.Draw3D(Var ARender:TRender);
 const
  interlace:array[0..3] of integer=(0,2,1,3);
 var
  ry:single;
  py:integer;
  x,y,i:integer;
 begin
 { for each pixel }
  ry:=-ARender.Height div 2;
  Ray.z:= 0;
  for i:=0 to 3 do begin
   py:=interlace[i];
   Ray.y:=ry+py;
   for y:=0 to (ARender.Height-1) div 4 do begin
    Ray.x:=-ARender.Width  div 2;
    for x:=0 to ARender.Width-1 do begin
    { find any intersection between the ray and an object }
     Pixel.Owner:=nil;
     Pixel.Dist:=BigZ;
     if FindIntersection(Eye,Ray,Pixel) then begin
     { object color }
      PixelColor(Eye,Pixel,1);
      ARender.PutPixel(x,py,Pixel.Color.R,Pixel.Color.G,Pixel.Color.B)
     end else
     { background color }
      ARender.PutPixel(x,py,Back.R,Back.G,Back.B);
     Ray.x:=Ray.x+1;
    end;
    Ray.y:=Ray.y+4;
    inc(py,4);
   end;
  end;
 end;

Function TScene.FindIntersection(Var Src,Dst:TPoint; Var APixel:TPixel):boolean;
 var
  obj:PPrimitive;
 begin
  FindIntersection:=False;
  obj:=Objects;
  while obj<>nil do begin
   if obj<>APixel.Owner then
   if obj^.Intersection(Src,Dst,APixel) then FindIntersection:=True;
   obj:=obj^.next;
  end;
 end;

Procedure TScene.PixelColor(Var Src:TPoint; Var APixel:TPixel;Generation:word);
 var
  lamp:PPrimitive;
  R,G,B:single;
  Transp:TPixel;

  procedure ComputeLamp;
   var
    light:TPoint;
    angle:single;
    fact :single;


    function ShadowLamp:boolean;
    {
      Light
        *
         \
          ##
         #### Shadow
          ##\
             \
              x
             Pixel
    }
     var
      Inter:TPixel;
     begin
      ShadowLamp:=false;
      Inter.Owner:=nil;
      Inter.Dist:=sqr(light.x-Pixel.Inter.x)+
                  sqr(light.x-Pixel.Inter.y)+
                  sqr(light.x-Pixel.Inter.z);
      if FindIntersection(Light,APixel.Inter,Inter) then begin
       if Inter.Owner<>APixel.Owner then begin
        ShadowLamp:=True;

        light.x:=Light.x-APixel.Inter.x;
        light.y:=Light.y-APixel.Inter.y;
        light.z:=Light.z-APixel.Inter.z;
        Normalize(Light);

        Angle:=NormalDot(light,APixel.Normal);
        if Angle>0 then begin
         fact:=Lamp^.Brightness*Inter.Owner^.Transparency*Angle/512;
         R:=R+(Lamp^.Color.R+Inter.Owner^.Color.R)*fact;
         G:=G+(Lamp^.Color.R+Inter.Owner^.Color.G)*fact;
         B:=B+(Lamp^.Color.R+Inter.Owner^.Color.B)*fact;
        end;

       end;
      end;
     end;

   begin
    if lamp^.BrightNess=0 then exit;

   { shadows }
    light:=lamp^.Center;
    if ShadowLamp then exit;

   { direct ligthning }
    light.x:=Light.x-APixel.Inter.x;
    light.y:=Light.y-APixel.Inter.y;
    light.z:=Light.z-APixel.Inter.z;
    Normalize(Light);

    Angle:=NormalDot(light,APixel.Normal);
    if Angle>0 then begin
     fact:=Lamp^.Brightness*Angle/256;
     R:=R+lamp^.color.r*fact;
     G:=G+lamp^.color.g*fact;
     B:=B+lamp^.color.b*fact;
    end;

   end;

 begin
  Normalize(APixel.Normal);
 {
    Light
      *      Eye
       \  :  /
        \ : /
         \:/
   -------+-------- (object color)
        pixel

 }
  R:=0;
  G:=0;
  B:=0;

  lamp:=lamps;
  while lamp<>nil do begin
   ComputeLamp;
   lamp:=lamp^.Next;
  end;

  lamp:=objects;
  while lamp<>nil do begin
   if lamp<>APixel.Owner then ComputeLamp;
   lamp:=lamp^.Next;
  end;

  R:=R*APixel.Owner^.Color.R;
  G:=G*APixel.Owner^.Color.G;
  B:=B*APixel.Owner^.Color.B;

  { Transparency }
  if (Generation<5)and(APixel.Owner^.Transparency>0) then begin
   Transp.Owner:=APixel.Owner;
   Transp.Dist:=BigZ;
   if FindIntersection(Src,APixel.Inter,Transp) then begin
    PixelColor(Src,Transp,Generation+1);
    R:=R+APixel.Owner^.Transparency*Transp.Color.R;
    G:=G+APixel.Owner^.Transparency*Transp.Color.G;
    B:=B+APixel.Owner^.Transparency*Transp.Color.B;
   end else begin
    R:=R+APixel.Owner^.Transparency*Back.R;
    G:=G+APixel.Owner^.Transparency*Back.G;
    B:=B+APixel.Owner^.Transparency*Back.B;
   end;
  end;

  if R>255 then R:=255; APixel.Color.R:=round(R);
  if G>255 then G:=255; APixel.Color.G:=round(G);
  if B>255 then B:=255; APixel.Color.B:=round(B);
 end;

Constructor TLamp.Create(Var AScene:TScene);
 begin
  Next:=AScene.Lamps;
  AScene.Lamps:=@Self;
  Center.x:=-150;
  Center.y:=0;
  Center.z:=-200;
  Radius:=5;
  Color.R:=255;
  Color.G:=255;
  Color.B:=255;
  Brightness:=1;
 end;

procedure TLamp.DrawXZ(Var ARender:TRender);
 begin
  inherited DrawXZ(ARender);
  ARender.Circle(ARender.Width div 2+Round(Center.X),
                 ARender.Height div 2-Round(Center.Z),
                 round(Radius),Color.R,Color.G,Color.B);
 end;

Constructor TSphere.Create(Var AScene:TScene);
 begin
  Next:=AScene.Objects;
  AScene.Objects:=@Self;

  Center.x:=0;
  Center.y:=0;
  Center.z:=0;
  Radius  :=100;

  Color.R:=255;
  Color.G:=0;
  Color.B:=0;

  Brightness:=0;
  Transparency:=0;
 end;

Function TSphere.Intersection(var Src,Dst:TPoint;Var Pixel:TPixel):boolean;
 Var
  Inter :TPoint;
  a,b,c :single;
  d,t   :single;
 Begin
  Intersection:=False;

 {           Src
             /
         Dst/
           x
          .Inter
         .
      --x-- object
      Pixel
    (t*Inter)
 }

  Inter.x:=Dst.x-Src.x;
  Inter.y:=Dst.y-Src.y;
  Inter.z:=Dst.z-Src.z;

 { solve the equation sphere-ray to find the intersections }
  a:=sqr(Inter.x)+sqr(Inter.y)+sqr(Inter.z);
  if a=0 then a:=0.001;

  b:=2*(Src.x-Center.x)*Inter.x+
     2*(Src.y-Center.y)*Inter.y+
     2*(Src.z-Center.z)*Inter.z;

  c:=Sqr(Src.x-Center.x)+
     Sqr(Src.y-Center.y)+
     Sqr(Src.z-Center.z)-
     Sqr(Radius);

  d:=Sqr(b)-4*a*c;

  if (d>=0) then begin
   { 1st Solution }
   t:=(-b-Sqrt(d))/(2*a);
   { 2nd Solution }
   if t<0 then t:=(-b+Sqrt(d))/(2*a);

   Inter.x:=t*Inter.x;
   Inter.y:=t*Inter.y;
   Inter.z:=t*Inter.z;

   d:=sqr(Inter.x)+sqr(Inter.y)+sqr(Inter.z);
   if d<Pixel.Dist then begin
    Pixel.Owner :=@Self;
    Pixel.Dist  :=d;
    Pixel.Point :=Inter;
    Pixel.Inter.x:=Src.x+Inter.x;
    Pixel.Inter.y:=Src.y+Inter.y;
    Pixel.Inter.z:=Src.z+Inter.z;
    Pixel.Normal.x:=Src.x+Inter.x-Center.x;
    Pixel.Normal.y:=Src.y+Inter.y-Center.y;
    Pixel.Normal.z:=Src.z+Inter.z-Center.z;
    Intersection:=True;
   end;

  end;

 end;

procedure TSphere.DrawXZ(Var ARender:TRender);
 begin
  inherited DrawXZ(ARender);
  ARender.Circle(ARender.Width div 2+Round(Center.X),
                 ARender.Height div 2-Round(Center.Z),
                 round(Radius),Color.R,Color.G,Color.B);
 end;

Var
 Scene :TScene;  { collection of 3D object }
 Lamp  :TLamp;   { a light source }
 Sphere:TSphere; { a 3D object }
 Render:TRender; { the screen  }

 s2:TSphere;
 s3:TSphere;

begin
 ShowVesaModes;
 Write('Press ENTER');
 Readln;

 Scene.Create;         { create a scene ...}
 Lamp.Create(Scene);

 Sphere.Create(Scene); { ... with a sphere ... }
 Sphere.Center.x:=80;

 s2.Create(Scene);
 s2.Center.x:=-10;
 s2.Center.y:=-40;
 s2.Radius:=80;
 s2.Color.R:=0;
 s2.Color.B:=128;
 s2.Transparency:=0.75;

 s3.Create(Scene);
 s3.Center.x:=+20;
 s3.Center.y:=+40;
 s3.Center.z:=-120;
 s3.Radius:=50;
 s3.Color.R:=0;
 s3.Color.G:=128;
 s3.Transparency:=0.75;

 Render.Create;         { create a render ... }
 (*
  Scene.DrawXZ(Render); { ... to display the scene }
  ReadLn;
 *)
  Scene.Draw3D(Render);
  Readln;
 Render.Destroy;       { Restore CRT Mode }
end.