Unit Render;
{
  Sample Raytracing program v2
  Copyright (C) 1998  Paul Toth

  TothPaul@multimania.com
  http://www.multimania.com/tothpaul

  Thanks to Cyrille FAVREAU for is Delphi RayTracer !
  http://www.multimania.com/cfavreau

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

Interface

Type
 TRender=object
  Width,Height:integer;
  Constructor Create;
  Destructor Destroy;
  function CenterX:integer;
  function CenterY:integer;
  procedure ResetBank;
  procedure PutPixel(x,y,r,g,b:word);
  procedure GetPixel(x,y:word; Var r,g,b:byte);
  procedure Clear;
  procedure Line(x1,y1,x2,y2,r,g,b:integer);
  procedure Circle(x,y,rd,r,g,b:integer);
  procedure SaveToFile(AFileName:string);
 end;

Implementation

{--- VESA ------------------------------------------------------}
Type
 TVesaInfo=record
  VESA         : array[1..4] of char;
  Version      : word;
  OEMName      : PChar;
  Capabilities : longint;
  VESAModes    : pointer;
  Memory       : word;
  Reserved     : array[20..255] of byte;
 end;

 TVesaMode=record
  Flags         :set of (fSupported,fMore,fBIOS,fColor,fGraphic);
  HiFlags       :byte;
  WindowAFlags  ,
  WindowBFlags  :set of (fAvail,fRead,fWrite);
  Granularity   :word;  { Bank*Granularity = Video Offset }
  WindowSize    :word;  { Allways 64K ? }
  WindowASegment:word;  { Allways $A000 ? }
  WindowBSegment:word;
  SetBank       :pointer;
  BytesPerline  :word;
 {-More-}
  Width         :word;
  Height        :word;
  CharWidth     :byte;
  CharHeight    :byte;
  PlaneCount    :byte;
  BitsPerPixel  :byte;
  BankCount     :byte;
  MemoryModel   :byte;
  BankSize      :byte;
  Reserved      :array[29..256] of byte;
 end;

Var
 VesaBank:pointer;
 VesaWin :word;
 VesaSeg :word;
 BankSize:word;
 LineSize:word;
 OldBank :word;

Function GetVesaInfo(Var VesaInfo:TVesaInfo):boolean; assembler;
 asm
  mov ax,4F00h
  les di,VesaInfo
  int 10h
  cmp al,004Fh
  je  @ok
   mov al,False
  @ok:
 end;

Function GetModeInfo(Mode:Word; Var VesaMode:TVesaMode):boolean; assembler;
  asm
   mov ax,$4F01
   mov cx,Mode
   les di,VesaMode
   int 10h
   cmp al,004Fh
   je  @ok
    mov al,False
   @Ok:
  end;

Procedure Error(Msg:string);
 begin
  writeln(msg);
  halt;
 end;

Function Hexa(w,size:word):string;
 const
  hex:array[0..$F] of char='0123456789ABCDEF';
 var
  result:string;
 begin
  result:='';
  repeat
   result:=hex[w and $F]+Result;
   w:=w shr 4;
  until w=0;
  while length(result)<size do result:='0'+result;
  Hexa:=Result;
 end;

Procedure ShowVesaModes;
 var
  VesaInfo:TVesaInfo;
  Mode:^Word;
  ModeInfo:TVesaMode;
 begin
  if not GetVesaInfo(VesaInfo) then Error('Not a VESA card');
  WriteLn(#16' ',VesaInfo.OEMName,' (Ver ',hi(VesaInfo.Version),'.',lo(VesaInfo.Version),') ',VesaInfo.Memory*64,'Ko');
  Mode:=VesaInfo.VesaModes;
  while mode^<>$FFFF do begin

   write(' mode ',hexa(mode^,3),'h at');

   If GetModeInfo(mode^,ModeInfo) then begin

    if fSupported in ModeInfo.Flags then begin

     if fAvail in ModeInfo.WindowAFlags then begin
      Write(' $',Hexa(ModeInfo.WindowASegment,4));
     end;

     if fAvail in ModeInfo.WindowBFlags then begin
      Write(' $',Hexa(ModeInfo.WindowASegment,4));
     end;

     Write(' Window=',ModeInfo.WindowSize,'K/',ModeInfo.Granularity:2,'K banks - ');

     if fGraphic in ModeInfo.Flags then write('Graphic ') else  write('Text    ');

     if fMore in ModeInfo.Flags then
      Write(ModeInfo.Width,'x',ModeInfo.Height,'x',longint(1) shl ModeInfo.BitsPerPixel)
     else
      if fColor in ModeInfo.Flags then write('color') else write('b&w');

     writeln;

    end else
     writeln('unsupported ???');

   end else
    WriteLn('unavailable ???');

   inc(mode);
  end;
 end;

Function SetMode(Mode:word):boolean; assembler;
 asm
  mov ax,$4F02
  mov bx,Mode
  int 10h
  cmp al,004Fh
  je @ok
   mov al,False
  @ok:
 end;

Function SetBank(ABank:word):boolean; assembler;
 asm
  mov  ax,ABank
  cmp  ax,OldBank
  je @Ok
  mov  OldBank,ax
  mov  bx,VesaWin
  mul  BankSize
  mov  dx,ax
  call VesaBank
  cmp  ax,004Fh
  je   @ok
   mov al,False
  @ok:
 end;

{--- TRender -----------------------------------------------}

Constructor TRender.Create;
 Const
  Mode=$112;
 var
  VesaInfo:TVesaInfo;
  VesaMode:TVesaMode;
 begin
  ShowVesaModes;
  Write('Press ENTER');
  Readln;

  if not GetVesaInfo(VesaInfo) then Error('Not a VESA card');
  if not GetModeInfo(Mode,VesaMode) then Error('Unsupported mode 112h');
  if VesaMode.BitsPerPixel<>24 then Error('Mode 112h isn''t a 16M color mode');

  Width :=VesaMode.Width;
  Height:=VesaMode.Height;

  VesaBank:=VesaMode.SetBank;
  BankSize:=64 div VesaMode.Granularity;
  LineSize:=VesaMode.BytesPerLine;

  if fAvail in VesaMode.WindowAFlags then begin
   VesaWin :=0;
   VesaSeg :=VesaMode.WindowASegment;
  end else begin
   if not (fAvail in VesaMode.WindowBFlags) then Error('Not VesaWin found ???');
   VesaWin :=1;
   VesaSeg :=VesaMode.WindowBSegment;
  end;

  if not SetMode(Mode) then Error('Unable to set VESA Mode 112h');

  OldBank:=$FFFF;
 end;

Destructor TRender.Destroy;
 begin
  asm mov ax,03h; int 10h end;
 end;

Function TRender.CenterX:integer;
 begin
  CenterX:=Width div 2;
 end;

Function TRender.CenterY:integer;
 begin
  CenterY:=Height div 2;
 end;

Procedure TRender.ResetBank;
 begin
  OldBank:=$FFFF;
 end;

Procedure TRender.Putpixel(x,y,r,g,b:word);
 var
  pixel :word;
 begin
  asm
   xor dx,dx
   mov ax,y
   mul LineSize
   mov bx,x
   mov cx,bx
   add bx,bx
   add bx,cx
   add ax,bx
   adc dx,0
   mov pixel,ax

   push dx
   call SetBank

   mov es,VesaSeg
   mov di,Pixel
   mov ax,B; stosb
   mov ax,G; stosb
   mov ax,R; stosb
  end;
 end;

Procedure TRender.Getpixel(x,y:word;Var r,g,b:byte);
 var
  pixel :word;
  _b,_r,_g:byte;
 begin
  asm
   xor dx,dx
   mov ax,y
   mul LineSize
   mov bx,x
   mov cx,bx
   add bx,bx
   add bx,cx
   add ax,bx
   adc dx,0
   mov pixel,ax

   push dx
   call SetBank

   mov es,VesaSeg
   mov di,Pixel
   mov al,es:[di]; inc di; mov _r,al
   mov al,es:[di]; inc di; mov _g,al
   mov al,es:[di]; inc di; mov _b,al
  end;
  b:=_b;
  r:=_r;
  g:=_g;
 end;

Procedure TRender.Clear;
 var
  x,y:word;
 begin
  for x:=0 to 14 { 3*640*480/64K } do begin
   SetBank(x);
   asm
    mov es,VesaSeg
    xor di,di
    mov cx,$FFFF
    mov al,0
    rep stosb
   end;
  end;
 end;

procedure TRender.Line(x1,y1,x2,y2,r,g,b:integer);
 var
  dx,dy:integer;
  ix,iy:integer;
  ex,ey:integer;
  ee,ii:integer;
 begin
  dx:=x2-x1; if dx<0 then begin dx:=-dx; ix:=-1 end else ix:=+1;
  dy:=y2-y1; if dy<0 then begin dy:=-dy; iy:=-1 end else iy:=+1;
  ex:=dx;
  ey:=dy;
  if ex<ey then ee:=ey else ee:=ex;
  for ii:=0 to ee do begin
   if (x1>=0)and(x1<Width)
   and(y1>=0)and(y1<Height) then PutPixel(x1,y1,r,g,b);
   inc(ex,dx);
   if ex>ee then begin
    dec(ex,ee);
    inc(x1,ix);
   end;
   inc(ey,dy);
   if ey>ee then begin
    dec(ey,ee);
    inc(y1,iy);
   end;
  end;
 end;

procedure TRender.Circle(x,y,rd,r,g,b:integer);
 var
  a:byte;
  ax,ay:integer;
  lx,ly:integer;
 begin
  lx:=rd;
  ly:=0;
  for a:=1 to 16 do begin
   ax:=round(rd*cos(a*PI/32));
   ay:=round(rd*sin(a*PI/32));
   line(x+lx,y+ly,x+ax,y+ay,r,g,b);
   line(x-lx,y+ly,x-ax,y+ay,r,g,b);
   line(x+lx,y-ly,x+ax,y-ay,r,g,b);
   line(x-lx,y-ly,x-ax,y-ay,r,g,b);
   lx:=ax;
   ly:=ay;
  end;
 end;

{--- 24Bits BMP -------------------------------------------------}

Type
 TBMPHeader=record

  bfType:array[1..2] of char;
  bfSize:LongInt;
  brReserved:LongInt;
  bgOffBits:LongInt;

  biSize:LongInt;
  biWidth:LongInt;
  biHeight:LongInt;
  biPlanes:word;
  biBitCount:word;
  biCompression:longint;
  biSizeImage:longint;
  biXPelsPerMeter:longint;
  biYPelsPerMeter:longint;
  biClrUsed:Longint;
  biClrImportant:longint;

 end;

procedure TRender.SaveToFile(AFileName:string);
 Var
  Header:TBMPHeader;
  f:file;
  ALine:array[0..639] of record R,G,B:byte; end;
  x,y:word;
 begin
  OldBank:=$FFFF;
  assign(f,AFileName);
  rewrite(f,1);
  Header.bfType:='BM';
  Header.bfSize:=640*480+SizeOf(Header);
  Header.brReserved:=0;
  Header.bgOffBits:=SizeOf(Header);
  Header.biSize:=40; { ??? }
  Header.biWidth:=640;
  Header.biHeight:=480;
  Header.biPlanes:=1;
  Header.biBitCount:=24;
  Header.biCompression:=0;
  Header.biSizeImage:=640*480;
  Header.biXPelsPerMeter:=3790; { ?? }
  Header.biYPelsPerMeter:=3800;
  Header.biClrUsed:=0;
  Header.biClrImportant:=0;
  BlockWrite(f,Header,SizeOf(Header));
  for y:=479 downto 0 do begin
   for x:=0 to 639 do begin
    with ALine[x] do GetPixel(x,y,r,g,b);
    PutPixel(x,y,0,0,0);
   end;
   BlockWrite(f,ALine,SizeOf(ALine));
  end;
  close(f);
 end;

end.