Unit Scenes;
{
  Sample Raytracing program v2
  Copyright (C) 1998  Paul Toth

  TothPaul@multimania.com
  http://www.multimania.com/tothpaul

  Thanks to Cyrille FAVREAU for is Delphi RayTracer !
  http://www.multimania.com/cfavreau

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

{$N+,E-}

Interface

Uses
 Render;

Type
 PPrimitive=^TPrimitive;

 TPoint=record x,y,z:single end;

 TRGBColor=record R,G,B:word end;
 TRGBLight=record R,G,B,Bright:single end;

 TPixel=record
  Owner :PPrimitive;
  Dist  :single;
  Color :TRGBColor;
  Point :TPoint;
  Inter :TPoint;
  Normal:TPoint;
  Reflection:TPoint;
 end;

 TPrimitive=object
  Next:PPrimitive;
  Center:TPoint;
  Radius:single;
  Color:TRGBColor;
  Brightness:single;
  Transparency:single;
  Reflection:single;
  function Intersection(var Src,Dst:TPoint;Var Pixel:TPixel):boolean; virtual;
  procedure DrawXZ(Var ARender:TRender); virtual;
 end;

 TScene=object
  Eye:TPoint;
{  Ray:TPoint; }
  Back:TRGBColor;
  Lamps:PPrimitive;
  Objects:PPrimitive;
{  Pixel:TPixel; }
  Generation:word;
  constructor Create;
  procedure DrawXZ(Var ARender:TRender);
  procedure RayXZ(Var ARender:TRender; Rx:integer);
  procedure Draw3D(Var ARender:TRender);
  Function FindIntersection(var Src,Dst:TPoint;Owner,Target:PPrimitive; var APixel:TPixel):boolean;
  procedure PixelColor(Var Src:TPoint;Var APixel:TPixel);
  function Light(Var Src:PPrimitive; Var APixel:TPixel;Var ALight:TRGBLight):boolean;
 end;

 TLamp=object(TPrimitive)
  constructor Create(Var AScene:TScene);
 end;

 TSphere=object(TPrimitive)
  constructor Create(Var AScene:TScene);
  function Intersection(var Src,Dst:TPoint;Var Pixel:TPixel):boolean; virtual;
 end;

Procedure Normalize(Var N:TPoint);
Function NormalDot(Var A,B:TPoint):single;

Implementation

Const
 BigZ=10000000;

Procedure Normalize(Var N:TPoint);
 var
  l:single;
 begin
  with N do begin
   l:=sqrt(sqr(x)+sqr(y)+sqr(z));
   if l<>0 then begin
    x:=x/l;
    y:=y/l;
    z:=z/l;
   end;
  end;
 end;

Function NormalDot(Var A,B:TPoint):single;
 begin
  NormalDot:=(A.X*B.X)+(A.Y*B.Y)+(A.Z*B.Z);
 end;

Procedure Reflect(APixel:TPixel;Var Reflex:TPoint);
{
 Point  Reflex
    \ : /
     \:/
  ----+----
    Inter

}
 var
  d:single;
 begin
  d:=-2*NormalDot(APixel.Point,APixel.Normal);
  Reflex.x:=APixel.Inter.x+APixel.Point.x+d*APixel.Normal.x;
  Reflex.y:=APixel.Inter.y+APixel.Point.y+d*APixel.Normal.y;
  Reflex.z:=APixel.Inter.z+APixel.Point.z+d*APixel.Normal.z;
 end;

{--- TPrimitive -----------------------------------------------------}
Function TPrimitive.Intersection(var Src,Dst:TPoint;Var Pixel:TPixel):boolean;
 begin
  Intersection:=false;
 end;

Procedure TPrimitive.DrawXZ(Var ARender:TRender);
 var
  cx,cy:integer;
 begin
  cx:=ARender.CenterX+Round(Center.X);
  cy:=ARender.CenterY-Round(Center.Z);
 { Center }
  ARender.Line(cx-5,cy,cx+5,cy,Color.R,Color.G,Color.B);
  ARender.Line(cx,cy-5,cx,cy+5,Color.R,Color.G,Color.B);
 { Radius }
  ARender.Circle(cx,cy,round(Radius),Color.R,Color.G,Color.B);
 end;

{--- TScene ---------------------------------------------------------}
Constructor TScene.Create;
 begin
  Objects:=nil;
  Lamps:=nil;

  Eye.x:=0;
  Eye.y:=0;
  Eye.z:=-800;

  Back.R:=64;
  Back.G:=64;
  Back.B:=64;
 end;

Procedure TScene.DrawXZ(Var ARender:TRender);
 var
  cx,cy:integer;
  obj:PPrimitive;
 begin
  cx:=ARender.CenterX;
  cy:=ARender.CenterY;
 { clear }
  ARender.Clear;
 { axes }
  ARender.Line(0,cy,ARender.Width-1,cy,0,0,255);
  ARender.Line(cx,0,cx,ARender.Height-1,0,0,255);
 { Camera }
  ARender.Line(0,cy,cx,cy-round(Eye.z),0,255,0);
  ARender.Line(ARender.Width-1,cy,cx,cy-Round(Eye.z),0,255,0);
 { Lamps }
  obj:=Lamps;
  while obj<>nil do begin
   obj^.DrawXZ(ARender);
   obj:=obj^.Next;
  end;
 { Objects }
  obj:=Objects;
  while obj<>nil do begin
   obj^.DrawXZ(ARender);
   obj:=obj^.Next;
  end;
 end;

Procedure TScene.RayXZ(Var ARender:TRender; Rx:integer);
 var
  cx,cy:integer;
  cl:TRGBColor;
  rd:single;
  Pixel:TPixel;
  Ray:TPoint;
  Reflex:TPoint;
  Reflected:TPixel;
 begin
  Generation:=1;

  cx:=ARender.CenterX;
  cy:=ARender.CenterY;
  Ray.Z:=0;
  Ray.Y:=0;
  Ray.X:=Rx;
  Pixel.Owner:=nil;
  Pixel.Dist:=BigZ;
  ARender.Line(cx+round(Eye.X),cy+round(-Eye.Z),cx+Rx,cy,255,255,255);
  if FindIntersection(Eye,Ray,nil,nil,Pixel) then begin

  { Normal (before normalization in PixelColor) }
   ARender.Line(cx+round(Pixel.Inter.X),
                cy-round(Pixel.Inter.Z),
                cx+round(Pixel.Inter.X+Pixel.Normal.X),
                cy-round(Pixel.Inter.Z+Pixel.Normal.Z),
                0,255,255);

   PixelColor(Eye,Pixel);
  { Ray }
   ARender.Line(cx+round(Eye.X),
                cy-round(Eye.Z),
                cx+round(Pixel.Inter.X),
                cy-round(Pixel.Inter.Z),
                Pixel.Color.R,Pixel.Color.G,Pixel.Color.B);

  { Reflection }
   if Pixel.Owner^.Reflection>0 then begin
    Reflect(Pixel,Reflex);

    ARender.Line(cx+round(Pixel.Inter.X),
                 cy-round(Pixel.Inter.Z),
                 cx+round(Reflex.X),
                 cy-round(Reflex.Z),
                 0,255,255);


    Reflected.Dist:=BigZ;
    if FindIntersection(Pixel.Inter,Reflex,Pixel.Owner,nil,Reflected) then begin
     ARender.Line(cx+round(Pixel.Inter.X),
                  cy-round(Pixel.Inter.Z),
                  cx+round(Reflected.Inter.X),
                  cy-round(Reflected.Inter.Z),
                  0,255,0);
    end;
   end;
  end;

 end;

Procedure TScene.Draw3D(Var ARender:TRender);
 const
  interlace:array[0..3] of integer=(0,2,1,3);
 var
  ry:single;
  py:integer;
  x,y,i:integer;
  Pixel:TPixel;
  Ray:TPoint;
 begin
 { for each pixel }
  ry:=-ARender.CenterY;
  Ray.z:= 0;
  for i:=0 to 3 do begin
   py:=interlace[i];
   Ray.y:=ry+py;
   for y:=0 to (ARender.Height-1) div 4 do begin
    Ray.x:=-ARender.CenterX;
    for x:=0 to ARender.Width-1 do begin
    { find any intersection between the ray and an object }
     Generation:=1;
     if (Ray.y=0)and(Ray.x=110) then begin
      Ray.x:=110;
     end;
     Pixel.Dist:=BigZ;
     if FindIntersection(Eye,Ray,nil,nil,Pixel) then begin
     { object color }
      PixelColor(Eye,Pixel);
      ARender.PutPixel(x,py,Pixel.Color.R,Pixel.Color.G,Pixel.Color.B)
     end else
     { background color }
      ARender.PutPixel(x,py,Back.R,Back.G,Back.B);
     Ray.x:=Ray.x+1;
    end;
    Ray.y:=Ray.y+4;
    inc(py,4);
   end;
  end;
 end;

Function TScene.FindIntersection(Var Src,Dst:TPoint; Owner,Target:PPrimitive; Var APixel:TPixel):boolean;
 var
  obj:PPrimitive;
 begin
  FindIntersection:=False;
  if Generation>10 then exit; inc(Generation);
  obj:=Objects;
  while obj<>nil do begin
   if (obj<>Owner)and(obj<>Target) then
   if obj^.Intersection(Src,Dst,APixel) then FindIntersection:=True;
   obj:=obj^.next;
  end;
 end;

Procedure TScene.PixelColor(Var Src:TPoint; Var APixel:TPixel);
 var
  ALight:TRGBLight;
  lamp:PPrimitive;
  R,G,B:single;

  fact :single;
  Reflex:TPoint;
  Reflected:TPixel;

  Transp:TPixel;
 begin
  Normalize(APixel.Normal);

  ALight.R:=0;
  ALight.G:=0;
  ALight.B:=0;
  ALight.Bright:=0;

 { Lamp light }
  lamp:=lamps;
  while lamp<>nil do begin
   Light(Lamp,APixel,ALight);
   lamp:=lamp^.Next;
  end;

 { objects light }
  lamp:=objects;
  while lamp<>nil do begin
   if lamp<>APixel.Owner then Light(Lamp,APixel,ALight);
   lamp:=lamp^.Next;
  end;

 { Total light }
  fact:=(ALight.Bright+APixel.Owner^.Brightness)*(1-APixel.Owner^.Transparency);
  R:=(APixel.Owner^.Color.R*Fact+ALight.R);
  G:=(APixel.Owner^.Color.G*Fact+ALight.G);
  B:=(APixel.Owner^.Color.B*Fact+ALight.B);

  if (Generation<10) then begin

  { Reflection }
   fact:=APixel.Owner^.Reflection;
   if Fact>0 then begin
    Reflect(APixel,Reflex);
    Reflected.Dist:=BigZ;
    if FindIntersection(APixel.Inter,Reflex,APixel.Owner,nil,Reflected) then begin
     PixelColor(APixel.Inter,Reflected);
     R:=Reflected.Color.R*fact+R*(1-fact);
     G:=Reflected.Color.G*fact+G*(1-fact);
     B:=Reflected.Color.B*fact+B*(1-fact);
    end;
   end;

  { Transparency }
   fact:=APixel.Owner^.Transparency;
   if fact>0 then begin
    Transp.Dist:=BigZ;
    if FindIntersection(Src,APixel.Inter,APixel.Owner,nil,Transp) then begin
     PixelColor(Src,Transp);
     R:=Transp.Color.R*fact+R*(1-fact);
     G:=Transp.Color.G*fact+G*(1-fact);
     B:=Transp.Color.B*fact+B*(1-fact);
    end else begin
     R:=R+Back.R*fact;
     G:=G+Back.G*fact;
     B:=B+Back.B*fact;
    end;
   end;

  end;

  if R>255 then R:=255; APixel.Color.R:=round(R);
  if G>255 then G:=255; APixel.Color.G:=round(G);
  if B>255 then B:=255; APixel.Color.B:=round(B);
 end;

Function TScene.Light(Var Src:PPrimitive; Var APixel:TPixel; Var ALight:TRGBLight):boolean;
 Var
  Step  :TRGBLight;
  Bright:single;
  Source:TPoint;
  NSource:TPoint;
  Shadow:TPixel;

  fact:single;
  Angle:single;
 {
  1- (Source)-->--(shadow)-----(shadow)-----(APixel)
  2- ( done ).....(Source)-->--(shadow)-----(APixel)
  3- ( done ).....( done ).....(Source)-->--(APixel)
 }

 begin
  Light:=False;

  Bright:=Src^.Brightness; { light source brightness }
  if Bright=0 then exit; { no light }

  Source:=Src^.Center;   { source of light }

  NSource.x:=Source.x-APixel.Inter.x; { lightning }
  NSource.y:=Source.y-APixel.Inter.y;
  NSource.z:=Source.z-APixel.Inter.z;
  Normalize(NSource);
  Angle:=NormalDot(NSource,APixel.Normal);
  if Angle<=0 then exit; { none }

  Step.R:=Src^.Color.R;  { light source color }
  Step.G:=Src^.Color.G;
  Step.B:=Src^.Color.B;

 { Shadow }
  Shadow.Owner:=Src;
  Shadow.Dist:=sqr(Source.x-APixel.Inter.x)+
               sqr(Source.y-APixel.Inter.y)+
               sqr(Source.z-APixel.Inter.z);
  while FindIntersection(Source,APixel.Inter,APixel.Owner,Shadow.Owner,Shadow) do begin
   fact:=Shadow.Owner^.Transparency;
   if fact=0 then exit;
   Step.R:=(Step.R*Bright+Shadow.Owner^.Color.R*fact);
   Step.G:=(Step.G*Bright+Shadow.Owner^.Color.G*fact);
   Step.B:=(Step.B*Bright+Shadow.Owner^.Color.B*fact);
   Bright:=Bright*fact;
   Source:=Shadow.Inter;
   Shadow.Dist:=sqr(Source.x-APixel.Inter.x)+
                sqr(Source.x-APixel.Inter.y)+
                sqr(Source.x-APixel.Inter.z);
  end;


 { Total Light }
  fact:=Bright*Angle;
  ALight.Bright:=ALight.Bright+fact;
  fact:=fact;
  ALight.R:=ALight.R+Step.R*fact;
  ALight.G:=ALight.G+Step.G*fact;
  ALight.B:=ALight.B+Step.B*fact;

  Light:=True;
 end;

{--- TLamp ----------------------------------------------------}
Constructor TLamp.Create(Var AScene:TScene);
 begin
  Next:=AScene.Lamps;
  AScene.Lamps:=@Self;
  Center.x:=-100;
  Center.y:=-100;
  Center.z:=-100;
  Radius:=5;
  Color.R:=255;
  Color.G:=255;
  Color.B:=255;
  Brightness:=1;
  Reflection:=0;
 end;

{--- TSphere -------------------------------------------------}

Constructor TSphere.Create(Var AScene:TScene);
 begin
  Next:=AScene.Objects;
  AScene.Objects:=@Self;

  Center.x:=0;
  Center.y:=0;
  Center.z:=0;
  Radius  :=100;

  Color.R:=255;
  Color.G:=255;
  Color.B:=255;

  Brightness:=0;
  Transparency:=0;
  Reflection:=0;
 end;

Function TSphere.Intersection(var Src,Dst:TPoint;Var Pixel:TPixel):boolean;
 Var
  Inter :TPoint;
  a,b,c :single;
  d,t   :single;
 Begin
  Intersection:=False;

  Inter.x:=Dst.x-Src.x;
  Inter.y:=Dst.y-Src.y;
  Inter.z:=Dst.z-Src.z;

 { solve the equation sphere-ray to find the intersections }
  a:=sqr(Inter.x)+sqr(Inter.y)+sqr(Inter.z);
  if a=0 then a:=0.001;

  b:=2*(Src.x-Center.x)*Inter.x+
     2*(Src.y-Center.y)*Inter.y+
     2*(Src.z-Center.z)*Inter.z;

  c:=Sqr(Src.x-Center.x)+
     Sqr(Src.y-Center.y)+
     Sqr(Src.z-Center.z)-
     Sqr(Radius);

  d:=Sqr(b)-4*a*c;

  if (d>0) then begin
   { 1st Solution }
   t:=(-b-Sqrt(d))/(2*a);
   { 2nd Solution }
   if t<0 then exit;{t:=(-b+Sqrt(d))/(2*a);}

   Inter.x:=t*Inter.x;
   Inter.y:=t*Inter.y;
   Inter.z:=t*Inter.z;

   d:=sqr(Inter.x)+sqr(Inter.y)+sqr(Inter.z);
   if d<Pixel.Dist then begin
    Pixel.Owner :=@Self;
    Pixel.Dist  :=d;
    Pixel.Point :=Inter;
    Inter.x:=Inter.x+Src.x;
    Inter.y:=Inter.y+Src.y;
    Inter.z:=Inter.z+Src.z;
    Pixel.Inter :=Inter;
    Inter.x:=Inter.x-Center.x;
    Inter.y:=Inter.y-Center.y;
    Inter.z:=Inter.z-Center.z;
    Pixel.Normal:=Inter;
    Intersection:=True;
   end;

  end;

 end;

end.